# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Derived class of server functions for Numpy/Scipy operations

    take a stream of assets which are raw acceleration values, process them, and return
    a stream of assets representing the results.

"""


__author__ = "Douglas Orr, Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2020 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import logging

import scipy.signal as sig
from scipy import fftpack
import numpy as np

from foglamp.common import logger, iprpc

G_ACCEL = 9.80665

_LOGGER = logger.setup(__name__, level=logging.INFO)


def bandwidth_filter_high(data, cutoff, order=10):
    """ High pass filter
    Args:
        data: signal data
        cutoff: cutoff frequency divided by Nyquist frequency
        order: Order of the butterworth filter
    Returns:
        filtered signal
    Raises:
    """
    _sos = sig.butter(order, cutoff, btype='high', output='sos')
    return sig.sosfiltfilt(_sos, data)


def bandwidth_filter_low(data, cutoff, order=10):
    """ Low pass filter
    Args:
        data: signal data
        cutoff: cutoff frequency divided by Nyquist frequency
        order: Order of the butterworth filter
    Returns:
        filtered signal
    Raises:
    """
    _sos = sig.butter(order, cutoff, btype='low', output='sos')
    return sig.sosfiltfilt(_sos, data)


def bandwidth_filter_bandpass(data, low_high_cutoff, order=10):
    """ Band pass filter
    Args:
        data: signal data
        low_high_cutoff: 2 element list with low & high cutoff frequency divided by Nyquist frequency
        order: Order of the butterworth filter
    Returns:
        filtered signal
    Raises:
    """
    _sos = sig.butter(order, low_high_cutoff, btype='bandpass', output='sos')
    return sig.sosfiltfilt(_sos, data)


def filter_signal(data, sampling_rate, f_lower, f_upper):
    """ Filters signal
    Args:
        data: signal data
        sampling_rate: sampling rate of signal
        f_lower: lower frequency of the filter
        f_upper: upper frequency of the filter
    Returns:
        filtered signal
    Raises:
    """
    nyq = sampling_rate / 2.0

    f_lower = max(0, f_lower)
    f_upper = min(sampling_rate/2, f_upper)

    if (f_lower == 0) and (f_upper == sampling_rate/2):
        data = data
    elif f_lower == 0:
        data = bandwidth_filter_low(data, f_upper/nyq)
    elif f_upper == sampling_rate/2:
        data = bandwidth_filter_high(data, f_lower/nyq)
    else:
        data = bandwidth_filter_bandpass(data, [f_lower/nyq, f_upper/nyq])

    return data


def hilbert(data):
    """ calculate absolute value of hilbert transform of signal
    Args:
        data: signal data
    Returns:
        absolute value of hilbert transform of signal
    Raises:
    """
    analytic_signal = sig.hilbert(data)
    return np.abs(analytic_signal)


class NPServer(iprpc.InterProcessRPC):
    """ Class for offloading numpy/scipy operations from filter plugin to a separate process
    """
    def __init__(self, service_name):
        """ Initialize numpy/scipy offloading server
        Args:
            service_name: name of the server
        Returns:
        Raises:
        """
        super().__init__()  # initialize rpc server

    def acc_to_velocity(self, signal, fs, calc_envelope=True, f_lower=4500, f_upper=8000):
        """ Processes raw acceleration data to calculate velocity, RMS acceleration/velocity & acceleration envelope
        Args:
            signal: signal data
            fs: sampling frequency of the signal
            calc_envelope: whether to calculate acceleration envelope
            f_lower,f_upper: frequency range of pass band to use during envelope calculation
        Returns:
            velocity: instantaneous velocity
            accelGRms: RMS acceleration in G
            accelGRmsPeak: Peak acceleration in G
            velGRms: RMS velocity in mm/s
            velGRmsPeak: Peak velocity in mm/s
            envelope: Acceleration envelope
        Raises:
        """
        signal = np.array(signal)
        accelGRms = np.sqrt(np.mean(signal ** 2)) / G_ACCEL
        accelGRmsPeak = np.max(signal) / G_ACCEL  # np.max(np.abs(signal))

        signal = sig.detrend(signal)

        amps = fftpack.fft(signal)
        freqs = fftpack.fftfreq(signal.shape[0], d=1/fs)
        fft_bin_width = freqs[1]
        skip_bins = int(round((fs/1000)/fft_bin_width))

        amps = amps / (2 * np.pi * (1j) * (freqs + [fft_bin_width/2]))
        amps[:skip_bins] = [0]
        amps[-skip_bins:] = [0]

        velocity = fftpack.ifft(amps)

        velocity = np.real(velocity)
        velocity = velocity * [1000]  # convert from m/s to mm/s

        velGRms = np.sqrt(np.mean(velocity ** 2))
        velGRmsPeak = np.max(velocity)  # np.max(np.abs(signal))

        if calc_envelope:
            envelope = hilbert(filter_signal(signal, fs, f_lower, f_upper))
            envelope = envelope / [G_ACCEL]
            envelope = envelope.tolist()
        else:
            envelope = None

        return velocity, accelGRms, accelGRmsPeak, velGRms, velGRmsPeak, envelope

    def plugin_shutdown(self):
        """ Shutdown numpy/scipy offloading server
        Args:
        Returns:
        Raises:
            SystemExit
        """
        _LOGGER.info("PLUGIN SHUTDOWN")
        raise SystemExit


if __name__ == "__main__":
    _LOGGER.error("SERVING...np functions")
    rpc = NPServer('np service')
    rpc.serve()
